from typing_extensions import Sequence
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal

from trading_sdk.core import Stream, PaginatedResponse, LogicError
from trading_sdk.market import (
  Market,
  Book,
  Order,
  OrderResponse,
  OrderState,
  Position,
  Rules,
  Trade,
)

from hyperliquid_sdk.core import wrap_exceptions

from .impl import (
  SpotMarketMixin,
  depth,
  depth_stream,
  spot_rules,
  open_orders,
  query_order,
  trades_history,
  trades_stream,
  spot_position,
  place_order,
  cancel_order,
)


@dataclass(frozen=True, kw_only=True)
class SpotMarket(SpotMarketMixin, Market):
  @property
  def venue_id(self) -> str:
    return 'hyperliquid'

  @property
  def exchange_id(self) -> str:
    return 'spot'

  @property
  def market_id(self) -> str:
    return f'{self.base_name}/{self.quote_name}:{self.asset_idx}'

  async def depth(self, *, levels: int | None = None) -> Book:
    return await depth(self)

  async def depth_stream(self, *, levels: int | None = None) -> Stream[Book]:
    return await depth_stream(self)

  async def rules(self, *, refetch: bool = False) -> Rules:
    return await spot_rules(self, refetch=refetch)

  async def open_orders(self) -> Sequence[OrderState]:
    return await open_orders(self)

  async def trades_stream(self) -> Stream[Trade]:
    return await trades_stream(self)

  async def position(self) -> Position:
    return await spot_position(self)

  @wrap_exceptions
  async def available_notional(self) -> Decimal:
    state = await self.client.info.spot_clearinghouse_state(self.address)
    for balance in state['balances']:
      if balance['token'] == self.quote_meta['index']:
        if balance['coin'] != self.quote_meta['name']:
          raise LogicError(f'Found balance with matching index {balance["token"]}, but wrong coin "{balance["coin"]}" != "{self.quote_name}"')
        total = Decimal(balance['total'])
        locked = Decimal(balance['hold'])
        return total - locked
    return Decimal(0)

  def trades_history(self, start: datetime, end: datetime) -> PaginatedResponse[Trade]:
    return PaginatedResponse(trades_history(self, start, end))

  @wrap_exceptions
  async def query_order(self, id: str) -> OrderState | None:
    return await query_order(self, id)

  @wrap_exceptions
  async def place_order(self, order: Order) -> OrderResponse:
    return await place_order(self, order)

  @wrap_exceptions
  async def cancel_order(self, id: str):
    return await cancel_order(self, id)



from .exc import wrap_exceptions
from .constants import (
  PRICE_MAX_DECIMALS,
  SPOT_PRICE_MAX_DECIMALS,
  FUTURES_PRICE_MAX_DECIMALS,
  MAX_SIGNIFICANT_FIGURES,
  MIN_ORDER_VALUE,
  MIN_RELATIVE_PRICE,
  MAX_RELATIVE_PRICE,
  round_price,
)
from .settings import Settings


from .impl import Settings
from .spot_exchange import SpotExchange
from .spot_market import SpotMarket
from .perps_exchange import PerpExchange
from .perps_market import PerpMarket
from .venue import HyperliquidMarket



from decimal import Decimal

from trading_sdk.market import PerpPosition

from hyperliquid_sdk.core import wrap_exceptions
from .mixin import PerpMarketMixin


@wrap_exceptions
async def position(self: PerpMarketMixin) -> PerpPosition:
  state = await self.client.info.clearinghouse_state(self.address, dex=self.dex_name)
  for entry in state["assetPositions"]:
    pos = entry["position"]
    if pos["coin"] == self.asset_name:
      return PerpPosition(size=Decimal(pos["szi"]), entry_price=Decimal(pos["entryPx"]))
  return PerpPosition()




from typing_extensions import AsyncIterable, Sequence, Any
from datetime import datetime
from decimal import Decimal

from trading_sdk.core import Stream
from trading_sdk.market import Trade

from hyperliquid_sdk.core import wrap_exceptions
from .mixin import SpotMarketMixin, PerpMarketMixin


def _parse_time(value: Any) -> datetime:
  from hyperliquid.core import timestamp as ts
  return ts.parse(value).astimezone()


@wrap_exceptions
async def trades_stream(self: SpotMarketMixin | PerpMarketMixin) -> Stream[Trade]:
  fills = await self.subscribe_user_fills()

  async def stream():
    async for chunk in fills:
      # Both spot and perps subscriptions aggregate by time.
      if not chunk.get("isSnapshot"):
        for f in (chunk.get("fills") or []):
          if f.get("coin") != self.asset_name:
            continue
          sign = 1 if f["side"] == "B" else -1
          yield Trade(
            id=str(f.get("tid")),
            price=Decimal(f["px"]),
            qty=Decimal(f["sz"]) * sign,
            time=_parse_time(f["time"]),
            maker=not f.get("crossed", False),
            fee=Trade.Fee(amount=Decimal(f["fee"]), asset=f["feeToken"]) if f.get("fee") is not None else None,
            details=f,
          )

  return Stream(stream(), fills.unsubscribe)


def trades_history(self: SpotMarketMixin | PerpMarketMixin, start: datetime, end: datetime) -> AsyncIterable[Sequence[Trade]]:
  async def gen():
    from hyperliquid.core import timestamp as ts

    start_ts, end_ts = ts.dump(start), ts.dump(end)
    async for chunk in self.client.info.user_fills_by_time_paged(self.address, start_ts, end_time=end_ts):
      trades: list[Trade] = []
      for f in chunk:
        if f.get("coin") != self.asset_name:
          continue
        sign = 1 if f["side"] == "B" else -1
        trades.append(
          Trade(
            id=str(f.get("tid")),
            price=Decimal(f["px"]),
            qty=Decimal(f["sz"]) * sign,
            time=ts.parse(f["time"]).astimezone(),
            maker=not f.get("crossed", False),
            fee=Trade.Fee(amount=Decimal(f["fee"]), asset=f["feeToken"]),
            details=f,
          )
        )
      if trades:
        yield trades

  return gen()




from dataclasses import dataclass

from trading_sdk import TradingVenue

from .impl import SharedMixin
from .perps_exchange import PerpExchange
from .spot_exchange import SpotExchange

@dataclass(frozen=True)
class HyperliquidMarket(SharedMixin, TradingVenue):
  @property
  def venue_id(self) -> str:
    return 'hyperliquid'

  async def exchange(self, exchange_id: str, /):
    if exchange_id == 'spot':
      return SpotExchange(shared=self.shared)
    else:
      return await self.perp_exchange(exchange_id)

  async def perp_exchange(self, exchange_id: str, /):
    return await PerpExchange.fetch(self.shared, dex=exchange_id or None)

  async def market(self, exchange_market_id: str, /):
    exchange_id, market_id = exchange_market_id.split(':', 1)
    exchange = await self.exchange(exchange_id)
    return await exchange.market(market_id)

  async def exchanges(self) -> list[TradingVenue.ExchangeDescription]:
    out: list[TradingVenue.ExchangeDescription] = [
      {'id': 'spot', 'type': 'spot'},
      {'id': '', 'type': 'perp'},
    ]
    dexs = await self.shared.load_perp_dexs()
    for dex in dexs.values():
      if dex is not None:
        out.append({'id': dex['name'], 'type': 'perp'})
    return out


from .mixin import Shared, SharedMixin, SpotMixin, PerpMixin, SpotMarketMixin, PerpMarketMixin, SpotMeta, PerpMeta, Settings

from .depth import depth, depth_stream
from .orders import open_orders, place_order, cancel_order, query_order
from .trades import trades_history, trades_stream

from .spot_rules import rules as spot_rules
from .spot_position import position as spot_position

from .perps_rules import rules as perps_rules
from .perps_position import position as perps_position

from .index import index
from .funding import next_funding, funding_history, funding_payments



from decimal import Decimal, ROUND_HALF_UP

PRICE_MAX_DECIMALS = 5
"""See https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/tick-and-lot-size"""
SPOT_PRICE_MAX_DECIMALS = 8
"""See https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/tick-and-lot-size"""
FUTURES_PRICE_MAX_DECIMALS = 6
"""See https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/tick-and-lot-size"""
MAX_SIGNIFICANT_FIGURES = 5
"""See https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/tick-and-lot-size"""
MIN_ORDER_VALUE = Decimal(10) # MIN ORDER VALUE IN USD
"""Not specified in the docs, but the API returns errors for orders with less than $10."""
MIN_RELATIVE_PRICE = Decimal('0.2')
"""Not specified in the docs, but the API returns errors for orders >80% away from the current price."""
MAX_RELATIVE_PRICE = Decimal('1.8')
"""Not specified in the docs, but the API returns errors for orders >80% away from the current price."""


def round_price(price: Decimal, max_sig_figs: int = MAX_SIGNIFICANT_FIGURES) -> Decimal:
  """
  Round `price` to at most `max_sig_figs` significant figures.
  - If `price` is zero or integral, return it unchanged.

  See https://hyperliquid.gitbook.io/hyperliquid-docs/for-developers/api/tick-and-lot-size for details
  """
  if price.is_zero():
    return price

  if price >= 10000:
    return price.to_integral_value()

  x = price.normalize()
  k = x.adjusted()

  # Number of decimal places needed to keep `max_sig_figs` significant digits
  decimal_places = max_sig_figs - 1 - k

  # quant = 10^(-decimal_places); works for positive or negative decimal_places
  quant = Decimal(1).scaleb(-decimal_places)

  return x.quantize(quant, rounding=ROUND_HALF_UP)


from functools import wraps
import inspect

from trading_sdk.core import NetworkError, ValidationError, ApiError, Error
from hyperliquid import core

def wrap_exceptions(fn):
  if inspect.iscoroutinefunction(fn):
    @wraps(fn)
    async def wrapper(*args, **kwargs): # type: ignore
      try:
        return await fn(*args, **kwargs)
      except core.NetworkError as e:
        raise NetworkError(*e.args) from e
      except core.ValidationError as e:
        raise ValidationError(*e.args) from e
      except core.ApiError as e:
        raise ApiError(*e.args) from e
      except core.Error as e:
        raise Error(*e.args) from e

  elif inspect.isgeneratorfunction(fn):
    @wraps(fn)
    async def wrapper(*args, **kwargs): # type: ignore
      try:
        return await fn(*args, **kwargs)
      except core.NetworkError as e:
        raise NetworkError(*e.args) from e
      except core.ValidationError as e:
        raise ValidationError(*e.args) from e
      except core.ApiError as e:
        raise ApiError(*e.args) from e
      except core.Error as e:
        raise Error(*e.args) from e
  else:
    @wraps(fn)
    def wrapper(*args, **kwargs):
      try:
        return fn(*args, **kwargs)
      except core.NetworkError as e:
        raise NetworkError(*e.args) from e
      except core.ValidationError as e:
        raise ValidationError(*e.args) from e
      except core.ApiError as e:
        raise ApiError(*e.args) from e
      except core.Error as e:
        raise Error(*e.args) from e
  return wrapper


from decimal import Decimal

from trading_sdk.market import Rules

from hyperliquid_sdk.core import (
  PRICE_MAX_DECIMALS,
  SPOT_PRICE_MAX_DECIMALS,
  MIN_ORDER_VALUE,
  MAX_RELATIVE_PRICE,
  MIN_RELATIVE_PRICE,
  wrap_exceptions
)

from .mixin import SpotMarketMixin

@wrap_exceptions
async def rules(self: SpotMarketMixin, *, refetch: bool = False) -> Rules:
  user_fees = await self.shared.load_user_fees(refetch=refetch)

  tick_decimals = min(
    PRICE_MAX_DECIMALS,
    SPOT_PRICE_MAX_DECIMALS - self.meta["base_meta"]["szDecimals"],
  )
  tick_size = Decimal(10) ** -tick_decimals

  lot_decimals = self.meta["base_meta"]["szDecimals"]
  lot_size = Decimal(10) ** -lot_decimals

  return Rules(
    base=self.base_name,
    quote=self.quote_name,
    fee_asset=self.quote_name,
    tick_size=tick_size,
    step_size=lot_size,
    min_value=MIN_ORDER_VALUE,
    rel_min_price=MIN_RELATIVE_PRICE,
    rel_max_price=MAX_RELATIVE_PRICE,
    maker_fee=Decimal(user_fees["userSpotAddRate"]),
    taker_fee=Decimal(user_fees["userSpotCrossRate"]),
    api=True,
    details={
      "user_fees": user_fees,
      "base_meta": self.meta["base_meta"],
      "quote_meta": self.meta["quote_meta"],
      "asset_meta": self.meta["asset_meta"],
    },
  )




from decimal import Decimal

from trading_sdk.core import Stream
from trading_sdk.market import Book

from hyperliquid_sdk.core import wrap_exceptions
from .mixin import SpotMarketMixin, PerpMarketMixin

Mixin = SpotMarketMixin | PerpMarketMixin

@wrap_exceptions
async def depth(self: Mixin) -> Book:
  book = await self.client.info.l2_book(self.asset_name)
  raw_bids, raw_asks = book["levels"]
  bids = [Book.Entry(price=Decimal(b["px"]), qty=Decimal(b["sz"])) for b in raw_bids]
  asks = [Book.Entry(price=Decimal(a["px"]), qty=Decimal(a["sz"])) for a in raw_asks]
  return Book(
    bids=sorted(bids, key=lambda e: e.price, reverse=True),
    asks=sorted(asks, key=lambda e: e.price),
  )


async def depth_stream(self: Mixin) -> Stream[Book]:
  l2 = await self.subscribe_l2_book(self.asset_name)

  async def stream():
    async for update in l2:
      raw_bids, raw_asks = update["levels"]
      # Hyperliquid `l2Book` is a snapshot-per-message feed.
      book = Book(
        bids=sorted(
          [Book.Entry(price=Decimal(b["px"]), qty=Decimal(b["sz"])) for b in raw_bids],
          key=lambda e: e.price,
          reverse=True,
        ),
        asks=sorted(
          [Book.Entry(price=Decimal(a["px"]), qty=Decimal(a["sz"])) for a in raw_asks],
          key=lambda e: e.price,
        ),
      )
      yield book

  return Stream(stream(), l2.unsubscribe)




from typing_extensions import TypedDict, Literal, NotRequired
from dataclasses import dataclass, field
import asyncio
import os

from trading_sdk.core import SDK, Stream, Subscription

from hyperliquid import Hyperliquid, Wallet
from hyperliquid.info.spot.spot_meta import SpotMetaResponse, SpotAssetInfo, SpotTokenInfo
from hyperliquid.info.methods.user_fees import UserFeesResponse
from hyperliquid.info.perps.perp_meta_and_asset_ctxs import (
  PerpMeta as PerpMetaResponse,
  PerpAssetCtx,
  PerpAssetInfo,
)
from hyperliquid.info.perps.perp_dexs import PerpDex
from hyperliquid.streams.user_fills import WsUserFills
from hyperliquid.streams.l2_book import L2BookData

from hyperliquid_sdk.core import Settings, wrap_exceptions

class DEX(TypedDict):
  name: str
  idx: int

class SpotMeta(TypedDict):
  asset_meta: SpotAssetInfo
  base_meta: SpotTokenInfo
  quote_meta: SpotTokenInfo

class PerpMeta(TypedDict):
  asset_idx: int
  asset_meta: PerpAssetInfo
  collateral_meta: SpotTokenInfo

def find_asset_idx(name: str, perp_meta: PerpMetaResponse) -> int:
  for idx, asset in enumerate(perp_meta['universe']):
    if asset['name'] == name:
      return idx
  raise ValueError(f'Perp {name} not found')

def find_dex_idx(name: str, dexs: list[PerpDex | None]) -> int:
  for idx, obj in enumerate(dexs):
    if obj and obj['name'] == name:
      return idx
  raise ValueError(f'DEX {name} not found')


def spot_meta_of(spot_index: int, /, *, spot_meta: SpotMetaResponse) -> SpotMeta:
  # Canonical spot id uses `spotMeta.universe[].index`.
  pos = next((i for i, a in enumerate(spot_meta['universe']) if a['index'] == spot_index), None)
  if pos is None:
    raise ValueError(f'Spot index {spot_index} not found in spot_meta universe')
  asset_meta = spot_meta['universe'][pos]
  base_idx, quote_idx = asset_meta['tokens']
  tokens_by_index = {t['index']: t for t in spot_meta['tokens']}
  return {
    'asset_meta': asset_meta,
    'base_meta': tokens_by_index[base_idx],
    'quote_meta': tokens_by_index[quote_idx],
  }


@dataclass(kw_only=True)
class Shared:
  client: Hyperliquid
  address: str
  settings: Settings = field(default_factory=Settings)

  # Cached meta (venue-wide).
  spot_meta: SpotMetaResponse | None = None
  # Lightweight DEX directory: idx -> PerpDex | None (wire type allows None entries).
  perp_dexs: dict[int, PerpDex | None] | None = None
  # Perp meta keyed by dex index; None key is used for the 'no dex' case.
  perp_metas: dict[int | None, PerpMetaResponse] = field(default_factory=dict)
  perp_asset_ctxs: dict[int | None, list[PerpAssetCtx]] = field(default_factory=dict)
  user_fees: UserFeesResponse | None = None

  # Stream subscriptions.
  user_fills_subscription: Subscription[WsUserFills] | None = None
  l2_book_subscriptions: dict[str, Subscription[L2BookData]] = field(default_factory=dict)

  # Locks for concurrent lazy loads.
  _spot_meta_lock: asyncio.Lock = field(default_factory=asyncio.Lock, init=False, repr=False)
  _perp_meta_lock: asyncio.Lock = field(default_factory=asyncio.Lock, init=False, repr=False)
  _fees_lock: asyncio.Lock = field(default_factory=asyncio.Lock, init=False, repr=False)

  @wrap_exceptions
  async def __aenter__(self):
    await self.client.__aenter__()
    return self

  @wrap_exceptions
  async def __aexit__(self, exc_type, exc_value, traceback):
    await self.client.__aexit__(exc_type, exc_value, traceback)

  @wrap_exceptions
  async def load_spot_meta(self, *, refetch: bool = False) -> SpotMetaResponse:
    if not refetch and self.spot_meta is not None:
      return self.spot_meta
    async with self._spot_meta_lock:
      if not refetch and self.spot_meta is not None:
        return self.spot_meta
      self.spot_meta = await self.client.info.spot_meta()
      return self.spot_meta

  @wrap_exceptions
  async def load_perp_dexs(self, *, refetch: bool = False) -> dict[int, PerpDex | None]:
    """
    Load and cache the list of DEXs (lightweight). Keyed by dex index.
    """
    if not refetch and self.perp_dexs is not None:
      return self.perp_dexs
    # No separate lock; cost is small and we typically call this rarely.
    dexs = await self.client.info.perp_dexs()
    self.perp_dexs = {idx: dex for idx, dex in enumerate(dexs)}
    return self.perp_dexs

  @wrap_exceptions
  async def load_perp_meta_for_dex(
    self,
    dex_name: str | None,
    *,
    refetch: bool = False,
  ) -> tuple[int, PerpMetaResponse, list[PerpAssetCtx]]:
    """
    Load perp meta + asset ctxs for a given dex name, caching per dex index.
    `dex_name = None` uses the default/no-dex universe.
    """
    if dex_name is None:
      dex_idx = 0
    else:
      dexs = await self.load_perp_dexs()
      dex_idx = find_dex_idx(dex_name, list(dexs.values()))

    key = dex_idx
    if not refetch and key in self.perp_metas and key in self.perp_asset_ctxs:
      return key, self.perp_metas[key], self.perp_asset_ctxs[key]
    async with self._perp_meta_lock:
      if not refetch and key in self.perp_metas and key in self.perp_asset_ctxs:
        return key, self.perp_metas[key], self.perp_asset_ctxs[key]
      perp_meta, asset_ctxs = await self.client.info.perp_meta_and_asset_ctxs(dex_name)
      self.perp_metas[key] = perp_meta
      self.perp_asset_ctxs[key] = asset_ctxs
      return key, perp_meta, asset_ctxs

  @wrap_exceptions
  async def load_user_fees(self, *, refetch: bool = False) -> UserFeesResponse:
    if not refetch and self.user_fees is not None:
      return self.user_fees
    async with self._fees_lock:
      if not refetch and self.user_fees is not None:
        return self.user_fees
      self.user_fees = await self.client.info.user_fees(self.address)
      return self.user_fees

  async def resolve_dex_idx(self, dex_name: str | None, *, refetch: bool = False) -> int:
    """
    Convenience wrapper that just returns the dex index for a given dex name.
    Uses the same caching as `load_perp_meta_for_dex`.
    """
    idx, _, _ = await self.load_perp_meta_for_dex(dex_name, refetch=refetch)
    return idx

  def user_fills_sub(self) -> Subscription[WsUserFills]:
    if self.user_fills_subscription is None:
      async def subscribe_user_fills() -> Stream[WsUserFills]:
        stream = await self.client.streams.user_fills(self.address, aggregate_by_time=True)
        return Stream(stream, stream.unsubscribe)
      self.user_fills_subscription = Subscription.of(subscribe_user_fills)
    return self.user_fills_subscription

  def l2_book_subscription(self, coin: str, /) -> Subscription[L2BookData]:
    if coin not in self.l2_book_subscriptions:
      async def subscribe() -> Stream[L2BookData]:
        stream = await self.client.streams.l2_book(coin)
        return Stream(stream, stream.unsubscribe)
      self.l2_book_subscriptions[coin] = Subscription.of(subscribe)
    return self.l2_book_subscriptions[coin]

  @wrap_exceptions
  async def spot_meta_of(
    self,
    spot_index: int,
    /,
    *,
    refetch: bool = False,
  ) -> SpotMeta:
    spot_meta = await self.load_spot_meta(refetch=refetch)
    return spot_meta_of(spot_index, spot_meta=spot_meta)

  @wrap_exceptions
  async def perp_meta_of(
    self,
    market: str,
    /,
    *,
    dex_name: str | None,
    refetch: bool = False,
  ) -> PerpMeta:
    spot_meta = await self.load_spot_meta(refetch=refetch)
    dex_idx, perp_meta, _ = await self.load_perp_meta_for_dex(dex_name, refetch=refetch)

    asset_idx = find_asset_idx(market, perp_meta)
    collateral_idx = perp_meta['collateralToken']
    tokens_by_index = {t['index']: t for t in spot_meta['tokens']}

    return PerpMeta(
      asset_idx=asset_idx,
      asset_meta=perp_meta['universe'][asset_idx],
      collateral_meta=tokens_by_index[collateral_idx],
    )

@dataclass(frozen=True)
class SharedMixin:
  shared: Shared

  @classmethod
  def http(
    cls, address: str | None = None, *, wallet: Wallet | None = None,
    mainnet: bool = True, settings: Settings = {}
  ):
    if address is None:
      address = os.environ['HYPERLIQUID_ADDRESS'] if mainnet else os.environ['HYPERLIQUID_TESTNET_ADDRESS']
    client = Hyperliquid.http(wallet, mainnet=mainnet, validate=settings.get('validate', True))
    return cls(shared=Shared(client=client, address=address, settings=settings))

  @classmethod
  def ws(
    cls, address: str | None = None, *, wallet: Wallet | None = None,
    mainnet: bool = True, settings: Settings = {}
  ):
    if address is None:
      address = os.environ['HYPERLIQUID_ADDRESS']
    client = Hyperliquid.ws(wallet, mainnet=mainnet, validate=settings.get('validate', True))
    return cls(shared=Shared(client=client, address=address, settings=settings))

  @property
  def client(self) -> Hyperliquid:
    return self.shared.client

  @property
  def address(self) -> str:
    return self.shared.address

  @property
  def settings(self) -> Settings:
    return self.shared.settings

  @property
  def index_price(self) -> Literal['oracle', 'mark']:
    return self.settings.get('index_price', 'oracle')

  async def __aenter__(self):
    await self.shared.__aenter__()
    return self

  async def __aexit__(self, exc_type, exc_value, traceback):
    await self.shared.__aexit__(exc_type, exc_value, traceback)

  async def subscribe_user_fills(self) -> Stream[WsUserFills]:
    return await self.shared.user_fills_sub().subscribe()

  async def subscribe_l2_book(self, coin: str, /) -> Stream[L2BookData]:
    return await self.shared.l2_book_subscription(coin).subscribe()


@dataclass(kw_only=True, frozen=True)
class SpotMixin(SharedMixin):
  ...


@dataclass(kw_only=True, frozen=True)
class SpotMarketMixin(SDK, SpotMixin):
  meta: SpotMeta

  @property
  def asset_idx(self) -> int:
    return self.meta['asset_meta']['index']

  @property
  def asset_meta(self) -> SpotAssetInfo:
    return self.meta['asset_meta']

  @property
  def asset_name(self) -> str:
    return self.asset_meta['name']

  @property
  def base_meta(self):
    return self.meta['base_meta']

  @property
  def quote_meta(self):
    return self.meta['quote_meta']

  @property
  def base_name(self) -> str:
    return self.meta['base_meta']['name']

  @property
  def quote_name(self) -> str:
    return self.meta['quote_meta']['name']

  @property
  def asset_id(self) -> int:
    return 10000 + self.asset_idx


@dataclass(kw_only=True, frozen=True)
class PerpMixin(SharedMixin):
  dex: DEX | None

  @classmethod
  async def fetch(cls, shared: Shared, *, dex: str | None = None):
    if dex: # we treat '' and None equivalently
      dex_idx = await shared.resolve_dex_idx(dex)
      dex_obj: DEX | None = {'name': dex, 'idx': dex_idx}
    else:
      dex_obj = None
    return cls(shared, dex=dex_obj)

  @property
  def dex_idx(self) -> int | None:
    if self.dex is not None:
      return self.dex['idx']

  @property
  def dex_name(self) -> str | None:
    if self.dex is not None:
      return self.dex['name']

@dataclass(kw_only=True, frozen=True)
class PerpMarketMixin(SDK, PerpMixin):
  meta: PerpMeta

  @property
  def asset_idx(self) -> int:
    return self.meta['asset_idx']

  @property
  def asset_meta(self) -> PerpAssetInfo:
    return self.meta['asset_meta']

  @property
  def asset_name(self) -> str:
    return self.asset_meta['name']

  @property
  def collateral_meta(self) -> SpotTokenInfo:
    return self.meta['collateral_meta']

  @property
  def collateral_name(self) -> str:
    return self.collateral_meta['name']

  @property
  def asset_id(self) -> int:
    # Per docs: base perps use asset index; builder perps use dex-scoped formula.
    if (dex := self.dex_idx) is None:
      return self.asset_idx
    return 100000 + dex * 10000 + self.asset_idx




from typing_extensions import Sequence
from dataclasses import dataclass

from trading_sdk.market import PerpExchange as _PerpExchange

from .impl import PerpMixin
from .perps_market import PerpMarket


@dataclass(frozen=True, kw_only=True)
class PerpExchange(PerpMixin, _PerpExchange):

  @property
  def venue_id(self) -> str:
    return 'hyperliquid'

  @property
  def exchange_id(self) -> str:
    return self.dex_name or ''

  async def markets(self) -> Sequence[str]:
    # Use default/no-dex universe here; callers that care about DEX can pass it when constructing.
    _, perp_meta, _ctxs = await self.shared.load_perp_meta_for_dex(self.dex_name)
    return [asset["name"] for asset in perp_meta["universe"]]

  async def market(self, market_id: str, /):
    meta = await self.shared.perp_meta_of(market_id, dex_name=self.dex_name)
    return PerpMarket(shared=self.shared, dex=self.dex, meta=meta)




from typing_extensions import TypedDict, Literal
from hyperliquid.exchange.order import TimeInForce

class Settings(TypedDict, total=False):
  validate: bool
  reduce_only: bool
  limit_tif: TimeInForce
  index_price: Literal['oracle', 'mark']


from typing_extensions import AsyncIterable, Sequence
from datetime import datetime, timedelta
from decimal import Decimal

from trading_sdk.market import FundingRate, FundingPayment

from hyperliquid.core import timestamp as ts
from hyperliquid_sdk.core import wrap_exceptions
from .mixin import PerpMarketMixin


@wrap_exceptions
async def next_funding(self: PerpMarketMixin) -> FundingRate:
  _, perp_meta, asset_ctxs = await self.shared.load_perp_meta_for_dex(self.dex_name, refetch=True)
  if perp_meta["universe"][self.asset_idx]["name"] != self.asset_name:
    raise ValueError(
      f"Expected asset {self.asset_name} at index {self.asset_idx}, got {perp_meta['universe'][self.asset_idx]['name']}"
    )

  funding = Decimal(asset_ctxs[self.asset_idx]["funding"])
  now = datetime.now().astimezone()
  next_time = now.replace(minute=0, second=0, microsecond=0) + timedelta(hours=1)
  return FundingRate(rate=funding, time=next_time)


@wrap_exceptions
async def funding_history(self: PerpMarketMixin, start: datetime, end: datetime) -> AsyncIterable[Sequence[FundingRate]]:
  start_ts, end_ts = ts.dump(start), ts.dump(end)
  async for chunk in self.client.info.funding_history_paged(self.asset_name, start_ts, end_time=end_ts):
    yield [
      FundingRate(rate=Decimal(entry["fundingRate"]), time=ts.parse(entry["time"]).astimezone())
      for entry in chunk
    ]


@wrap_exceptions
async def funding_payments(self: PerpMarketMixin, start: datetime, end: datetime) -> AsyncIterable[Sequence[FundingPayment]]:
  start_ts, end_ts = ts.dump(start), ts.dump(end)
  async for chunk in self.client.info.user_funding_paged(self.address, start_ts, end_time=end_ts):
    payments: list[FundingPayment] = []
    for p in chunk:
      if p["delta"]["coin"] != self.asset_name:
        continue
      t = ts.parse(p["time"]).astimezone()
      if t < start or t > end:
        continue
      payments.append(FundingPayment(amount=Decimal(p["delta"]["usdc"]), time=t))
    if payments:
      yield payments




from decimal import Decimal

from trading_sdk.market import Position

from hyperliquid_sdk.core import wrap_exceptions
from .mixin import SpotMarketMixin


@wrap_exceptions
async def position(self: SpotMarketMixin) -> Position:
  state = await self.client.info.spot_clearinghouse_state(self.address)
  for balance in state["balances"]:
    if balance["token"] == self.meta["base_meta"]["index"]:
      total = Decimal(balance["total"])
      return Position(size=total)
  return Position()




from decimal import Decimal

from hyperliquid_sdk.core import wrap_exceptions
from .mixin import PerpMarketMixin


@wrap_exceptions
async def index(self: PerpMarketMixin) -> Decimal:
  _, perp_meta, asset_ctxs = await self.shared.load_perp_meta_for_dex(self.dex_name, refetch=True)
  if perp_meta["universe"][self.asset_idx]["name"] != self.asset_name:
    raise ValueError(
      f"Expected asset {self.asset_name} at index {self.asset_idx}, got {perp_meta['universe'][self.asset_idx]['name']}"
    )

  ctx = asset_ctxs[self.asset_idx]
  if self.index_price == "oracle" or (mark := ctx.get("markPx")) is None:
    return Decimal(ctx["oraclePx"])
  return Decimal(mark)




from typing_extensions import Sequence
from dataclasses import dataclass

from trading_sdk.market import Exchange as _Exchange

from .impl import SpotMixin
from .spot_market import SpotMarket


def _market_id(base: str, quote: str, asset_idx: int) -> str:
  return f'{base}/{quote}:{asset_idx}'


def parse_market_id(market_id: str) -> tuple[str, str, int]:
  """Parse canonical spot `market_id` format: `BASE/QUOTE:ASSET_IDX`."""
  ticker, idx_str = market_id.rsplit(':', 1)
  base, quote = ticker.split('/', 1)
  return base, quote, int(idx_str)


@dataclass(frozen=True, kw_only=True)
class SpotExchange(SpotMixin, _Exchange):

  @property
  def venue_id(self) -> str:
    return 'hyperliquid'

  @property
  def exchange_id(self) -> str:
    return 'spot'

  async def markets(self) -> Sequence[str]:
    meta = await self.shared.load_spot_meta()
    out: list[str] = []
    for asset in meta['universe']:
      spot_index = asset['index']
      m = await self.shared.spot_meta_of(spot_index)
      out.append(_market_id(m['base_meta']['name'], m['quote_meta']['name'], spot_index))
    return out

  async def market(self, market_id: str, /):
    base, quote, spot_index = parse_market_id(market_id)
    meta = await self.shared.spot_meta_of(spot_index)
    if meta['base_meta']['name'] != base or meta['quote_meta']['name'] != quote:
      raise ValueError(
        f"Spot market_id mismatch for idx={spot_index}: expected {base}/{quote}, got {meta['base_meta']['name']}/{meta['quote_meta']['name']}"
      )
    return SpotMarket(shared=self.shared, meta=meta)




from typing_extensions import Any, Sequence
from decimal import Decimal

from trading_sdk.core import ApiError
from trading_sdk.market import Order, OrderResponse, OrderState
from trading_sdk.util import fmt_num

from hyperliquid.exchange.cancel import Cancel as CancelWire
from hyperliquid.exchange.order import Order as OrderWire
from hyperliquid.info.methods.order_status import OrderStatusResponse

from hyperliquid_sdk.core import round_price, wrap_exceptions
from .mixin import SpotMarketMixin, PerpMarketMixin


def _active(status: str) -> bool:
  # conservative: only 'open' is definitely active; 'triggered' is still live too
  return status in {"open", "triggered", "scheduledCancel"}


def _export_order(self: SpotMarketMixin | PerpMarketMixin, o: Order) -> OrderWire:
  if o["type"] == "LIMIT":
    tif = self.settings.get("limit_tif", "Gtc")
  elif o["type"] == "POST_ONLY":
    tif = "Alo"
  else:
    raise ValueError(f"Unknown order type: {o['type']}")

  qty = Decimal(o["qty"])
  price = round_price(Decimal(o["price"]))
  return {
    "a": self.asset_id,
    "b": qty >= 0,
    "p": fmt_num(price),
    "s": fmt_num(abs(qty)),
    "r": self.settings.get("reduce_only", False),
    "t": {"limit": {"tif": tif}},
  }


@wrap_exceptions
async def open_orders(self: SpotMarketMixin | PerpMarketMixin) -> Sequence[OrderState]:
  dex = getattr(self, "dex", None)
  if dex is not None:
    orders = await self.client.info.open_orders(self.address, dex=dex)
  else:
    orders = await self.client.info.open_orders(self.address)

  out: list[OrderState] = []
  for o in orders:
    if o.get("coin") != self.asset_name:
      continue
    qty = Decimal(o["origSz"])
    out.append(
      OrderState(
        id=str(o["oid"]),
        price=Decimal(o["limitPx"]),
        qty=qty,
        filled_qty=qty - Decimal(o["sz"]),
        active=True,
        details=o,
      )
    )
  return out


@wrap_exceptions
async def place_order(self: SpotMarketMixin | PerpMarketMixin, order: Order) -> OrderResponse:
  wire = _export_order(self, order)
  result = await self.client.exchange.order(wire)
  if result["status"] != "ok":
    raise ApiError(result)

  statuses = result["response"]["data"]["statuses"]
  if not statuses:
    raise ApiError({"error": "empty status list", "details": result})

  s = statuses[0]
  if (err := s.get("error")) is not None:
    raise ApiError(err)
  if (resting := s.get("resting")) is not None:
    return OrderResponse(id=str(resting["oid"]), details=s)
  if (filled := s.get("filled")) is not None:
    return OrderResponse(id=str(filled["oid"]), details=s)
  raise ApiError({"error": "unknown order status", "details": s})


@wrap_exceptions
async def cancel_order(self: SpotMarketMixin | PerpMarketMixin, id: str) -> Any:
  cancel: CancelWire = {"a": self.asset_id, "o": int(id)}
  result = await self.client.exchange.cancel(cancel)
  if result["status"] != "ok":
    raise ApiError(result)
  statuses = result["response"]["data"]["statuses"]
  if not statuses:
    raise ApiError({"error": "empty status list", "details": result})
  s = statuses[0]
  if s == "success":
    return s
  if isinstance(s, dict) and (err := s.get("error")) is not None:
    raise ApiError(err)
  raise ApiError({"error": "unknown cancel status", "details": s})


@wrap_exceptions
async def query_order(self: SpotMarketMixin | PerpMarketMixin, id: str) -> OrderState | None:
  status: OrderStatusResponse = await self.client.info.order_status(self.address, int(id))
  if status["status"] == "unknownOid":
    return None

  entry = status["order"]
  o = entry["order"]
  if o.get("coin") != self.asset_name:
    return None

  qty = Decimal(o["origSz"])
  return OrderState(
    id=str(o["oid"]),
    price=Decimal(o["limitPx"]),
    qty=qty,
    filled_qty=qty - Decimal(o["sz"]),
    active=_active(entry["status"]),
    details=status,
  )




from typing_extensions import AsyncIterable, Sequence
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal

from trading_sdk.core import Stream, PaginatedResponse, LogicError
from trading_sdk.market import (
  PerpMarket as _PerpMarket,
  Book,
  Order,
  OrderResponse,
  OrderState,
  PerpPosition,
  Rules,
  Trade,
  FundingRate,
  FundingPayment,
)

from hyperliquid_sdk.core import wrap_exceptions

from .impl import (
  PerpMarketMixin,
  depth,
  depth_stream,
  perps_rules,
  index,
  next_funding,
  funding_history,
  funding_payments,
  perps_position,
  open_orders,
  place_order,
  cancel_order,
  query_order,
  trades_history,
  trades_stream,
)


@dataclass(frozen=True, kw_only=True)
class PerpMarket(PerpMarketMixin, _PerpMarket):
  @property
  def venue_id(self) -> str:
    return "hyperliquid"

  @property
  def exchange_id(self) -> str:
    return self.dex_name or ''

  @property
  def market_id(self) -> str:
    return self.asset_name

  @wrap_exceptions
  async def depth(self, *, levels: int | None = None) -> Book:
    return await depth(self)

  async def depth_stream(self, *, levels: int | None = None) -> Stream[Book]:
    return await depth_stream(self)

  @wrap_exceptions
  async def rules(self, *, refetch: bool = False) -> Rules:
    return await perps_rules(self, refetch=refetch)

  @PaginatedResponse.lift
  def trades_history(self, start: datetime, end: datetime) -> AsyncIterable[Sequence[Trade]]:
    return trades_history(self, start, end)

  async def open_orders(self) -> Sequence[OrderState]:
    return await open_orders(self)

  async def trades_stream(self) -> Stream[Trade]:
    return await trades_stream(self)

  async def perp_position(self) -> PerpPosition:
    return await perps_position(self)

  @wrap_exceptions
  async def available_notional(self) -> Decimal:
    state = await self.client.info.spot_clearinghouse_state(self.address)
    for balance in state['balances']:
      if balance['token'] == self.collateral_meta['index']:
        if balance['coin'] != self.collateral_name:
          raise LogicError(f'Found balance with matching index {balance["token"]}, but wrong coin "{balance["coin"]}" != "{self.collateral_name}"')
        total = Decimal(balance['total'])
        locked = Decimal(balance['hold'])
        collateral = total - locked
        leverage = self.asset_meta['maxLeverage']
        return collateral * leverage

    return Decimal(0)

  async def place_order(self, order: Order) -> OrderResponse:
    return await place_order(self, order)

  async def query_order(self, id: str) -> OrderState | None:
    return await query_order(self, id)

  async def cancel_order(self, id: str):
    return await cancel_order(self, id)

  async def index(self) -> Decimal:
    return await index(self)

  async def next_funding(self) -> FundingRate:
    return await next_funding(self)

  def funding_history(self, start: datetime, end: datetime) -> PaginatedResponse[FundingRate]:
    return PaginatedResponse(funding_history(self, start, end))

  def funding_payments(self, start: datetime, end: datetime) -> PaginatedResponse[FundingPayment]:
    return PaginatedResponse(funding_payments(self, start, end))
